/******************************************************************************
 *
 * Filename: disasm.c
 *
 * Author:   Robert Thornburrow
 * Project:  JavaVM
 * Created:  19th January 1998
 *
 * Purpose:  Class file disassembler.
 *
 * (c) Copyright R.H.Thornburrow, 1998
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *****************************************************************************/

#include <stdio.h>
#include <string.h>

#include "config.h"
#include "classfile.h"
#include "instructions.h"

/* level:
 * 1 = method code
 * 2 = constant pool table
 * 4 = class attributes
 * 8 = method attributes
 */

#define PEEKU2(code,n) ( (code[(n)+1] << 8) | code[(n)+2] )
#define PEEKU4(code,n) ( (code[(n)+1] << 24) | (code[(n)+2] << 16) | (code[(n)+3] << 8) | code[(n)+4] )

#define DECLARE_PRINTF() int len, tlen = 0;
#define PRINTF(x) len = sprintf( buf, x ); buf += len; tlen += len;
#define PRINTF2(x1,x2) len = sprintf( buf, x1, x2 ); buf += len; tlen += len;
#define PRINTF3(x1,x2,x3) len = sprintf( buf, x1, x2, x3 ); buf += len; tlen += len;
#define PUTCHAR(x) *buf = x; buf++; tlen++;
#define RETURN_PRINTF() return tlen;

#ifdef _DEBUG

u2 getU2( u2 v )
{
#ifdef WIN32
	u1* b = (u1*) &v;
	return (b[0] << 8) | b[1];
#else
	return v;
#endif
}

/*u4 getU4( u4 v )
{
#ifdef WIN32
	u1* b = (u1*) &v;
	return (b[0] << 24) | (b[1] << 16) | (b[2] << 8) | b[3];
#else
	return v;
#endif
}*/

int sprintTypeName( char* buf, u1 t )
{
	DECLARE_PRINTF();

	switch( t )
	{
	case T_ANY:		PRINTF( "any" ); break;
	case T_OBJREF:	PRINTF( "objref" ); break;
	case T_ARRAY:	PRINTF( "array" ); break;

	case T_BOOLEAN: PRINTF( "boolean" ); break;
	case T_CHAR:	PRINTF( "char" ); break;
	case T_FLOAT:	PRINTF( "float" ); break;
	case T_DOUBLE:	PRINTF( "double" ); break;
	case T_BYTE:	PRINTF( "byte" ); break;
	case T_SHORT:	PRINTF( "short" ); break;
	case T_INT:		PRINTF( "int" ); break;
	case T_LONG:	PRINTF( "long" ); break;
	default:		PRINTF2( "<unknown 0x%x>", t ); break;
	}

	RETURN_PRINTF();
}


int sprintLiteral( char* buf, ClassFile* cf, u2 index )
{
	DECLARE_PRINTF();

	if( cf && (index > 0 )
		&& (index < cf->constant_pool_count)
		&& cf->constant_pool[index] )
	{
		cp_info* cp = cf->constant_pool[index];

		switch( cp->tag )
		{
		case CONSTANT_Utf8:
			break;
		case CONSTANT_Integer:
			PRINTF2( "Integer( %d )", ((int) ((CONSTANT_Integer_info*) cp)->bytes) );
			break;
		case CONSTANT_Float:
			PRINTF2( "Float( %f )", ((double) ((CONSTANT_Float_info*) cp)->bytes) );
			break;
		case CONSTANT_Long:
			PRINTF( "Long( ? )" );
			break;
		case CONSTANT_Double:
			PRINTF( "Double( ? )" );
			break;
		case CONSTANT_String:
			if( CHECKCONSTANT( cf, ((CONSTANT_String_info*) cp)->string_index, CONSTANT_Utf8 ) )
			{
				CONSTANT_Utf8_info* utf8 = (CONSTANT_Utf8_info*) cf->constant_pool[ ((CONSTANT_String_info*) cp)->string_index ];

				PRINTF( "String ( \"" );
				strncpy( buf, (const char*) utf8->bytes, utf8->length );
				buf += utf8->length; tlen += utf8->length;
				PRINTF( "\" )" );
			}
			else
				PRINTF( "String( ? )" );
			break;
		}
	}

	RETURN_PRINTF();
}


void writeUTF8( u1* bytes, u2 count )
{
	u2 n;

	for( n = 0; n < count; n++ )
	{
		if(( bytes[n] >= 0x01 ) && ( bytes[n] <= 0x7f ))
		{
			fputc( bytes[n], stdout );
		}
		else if( (bytes[n] & 0x0e) == 0x0c )
		{
			fputc( ((bytes[n] & 0x1f) << 6) + (bytes[n+1] & 0x3f), stdout );
			n++;
		}
		else if( (bytes[n] & 0x0e) == 0x0e )
		{
			fputc( ((bytes[n] & 0xf) << 12) + ((bytes[n+1] & 0x3f) << 6) + (bytes[n+2] & 0x3f), stdout );
			n+=2;
		}
	}
}

#endif /* _DEBUG */

int sprintDescriptor( char* buf, ClassFile* cf, u2 index )
{
	DECLARE_PRINTF();

	if( CHECKCONSTANT( cf, index, CONSTANT_Utf8 ) )
	{
		CONSTANT_Utf8_info* cu = (CONSTANT_Utf8_info*) cf->constant_pool[index];
		u2 n, array = 0;
		
		for( n = 0; n < cu->length; n++ )
		{
			switch( cu->bytes[n] )
			{
			case 'B': PRINTF( "byte" ); break;
			case 'C': PRINTF( "char" ); break;
			case 'D': PRINTF( "double" ); break;
			case 'F': PRINTF( "float" ); break;
			case 'I': PRINTF( "int" ); break;
			case 'J': PRINTF( "long" ); break;
			case 'L':
				for( n++; (n < cu->length) && (cu->bytes[n] != ';'); n++ )
				{
					if( cu->bytes[n] == '/' )
					{
						PUTCHAR( '.' );
					}
					else if(( cu->bytes[n] >= 0x01 ) && ( cu->bytes[n] <= 0x7f ))
					{
						PUTCHAR( cu->bytes[n] );
					}
				}
				break;
			case 'S': PRINTF( "short" ); break;
			case 'Z': PRINTF( "boolean" ); break;
			case '[': array++;
			}
		}

		for( n = 0; n < array; n++ )
		{
			PRINTF( "[]" );
		}
	}

	RETURN_PRINTF();
}

int sprintMethodParameters( char* buf, ClassFile* cf, u2 index )
{
	DECLARE_PRINTF();

	if( CHECKCONSTANT( cf, index, CONSTANT_Utf8 ) )
	{
		CONSTANT_Utf8_info* cu = (CONSTANT_Utf8_info*) cf->constant_pool[index];
		u2 n, m, array = 0;
		
		if( !cu->length || (cu->bytes[0] != '(') )
			return 0;

		PRINTF( "(" );

#define SHOW_ARRAY for( m = 0; m < array; m++ ) { PRINTF( "[]" ); }; array = 0;

		for( n = 1; (n < cu->length) && (cu->bytes[n] != ')'); n++ )
		{
			switch( cu->bytes[n] )
			{
			case 'B': PRINTF( " byte" ); SHOW_ARRAY; break;
			case 'C': PRINTF( " char" ); SHOW_ARRAY; break;
			case 'D': PRINTF( " double" ); SHOW_ARRAY; break;
			case 'F': PRINTF( " float" ); SHOW_ARRAY; break;
			case 'I': PRINTF( " int" ); SHOW_ARRAY; break;
			case 'J': PRINTF( " long" ); SHOW_ARRAY; break;
			case 'L':
				PRINTF( " " );
				for( n++; (n < cu->length) && (cu->bytes[n] != ';'); n++ )
				{
					if( cu->bytes[n] == '/' )
					{
						PUTCHAR( '.' );
					}
					else if(( cu->bytes[n] >= 0x01 ) && ( cu->bytes[n] <= 0x7f ))
					{
						PUTCHAR( cu->bytes[n] );
					}
				}
				SHOW_ARRAY;
				break;
			case 'S': PRINTF( " short" ); SHOW_ARRAY; break;
			case 'Z': PRINTF( " boolean" ); SHOW_ARRAY; break;
			case '[': array++; continue;
			}

			if( (n+1 < cu->length) && (cu->bytes[n+1] != ')' ) )
			{
				PRINTF( "," );
			}
			else
			{
				PRINTF( " " );
			}
		}

		PRINTF( ")" );
	}

	RETURN_PRINTF();
}

int sprintMethodReturn( char* buf, ClassFile* cf, u2 index )
{
	DECLARE_PRINTF();

	if( CHECKCONSTANT( cf, index, CONSTANT_Utf8 ) )
	{
		CONSTANT_Utf8_info* cu = (CONSTANT_Utf8_info*) cf->constant_pool[index];
		u2 n, array = 0;

		for( n = 0; (n < cu->length) && (cu->bytes[n] != ')'); n++);
		
		for( ; n < cu->length; n++ )
		{
			switch( cu->bytes[n] )
			{
			case 'B': PRINTF( "byte" ); break;
			case 'C': PRINTF( "char" ); break;
			case 'D': PRINTF( "double" ); break;
			case 'F': PRINTF( "float" ); break;
			case 'I': PRINTF( "int" ); break;
			case 'J': PRINTF( "long" ); break;
			case 'L':
				for( n++; (n < cu->length) && (cu->bytes[n] != ';'); n++ )
				{
					if( cu->bytes[n] == '/' )
					{
						PUTCHAR( '.' );
					}
					else if(( cu->bytes[n] >= 0x01 ) && ( cu->bytes[n] <= 0x7f ))
					{
						PUTCHAR( cu->bytes[n] );
					}
				}
				break;
			case 'S': PRINTF( "short" ); break;
			case 'V': PRINTF( "void" ); break;
			case 'Z': PRINTF( "boolean" ); break;
			case '[': array++;
			}
		}

		for( n = 0; n < array; n++ )
		{
			PRINTF( "[]" );
		}

		if( cu->bytes[cu->length-1] != ')' )
		{
			PRINTF( " " );
		}
	}

	RETURN_PRINTF();
}

int checkSimpleName( ClassFile* cf, u2 index, const char* compare )
{
	if( CHECKCONSTANT( cf, index, CONSTANT_Utf8 ) )
	{
		CONSTANT_Utf8_info* cu = (CONSTANT_Utf8_info*) cf->constant_pool[index];
		
		/*{
			char buf[255];
			strncpy( buf, (const char*) cu->bytes, cu->length );
			buf[cu->length] = '\0';
			printf( "Comparing \"%s\" to \"%s\"\n", buf, compare );
		}*/

		if( strlen( compare ) != cu->length )
			return 0;

		if( strncmp( (const char*) cu->bytes, compare, cu->length ) == 0 )
			return 1;
	}

	return 0;
}

int sprintSimpleName( char* buf, ClassFile* cf, u2 index )
{
	u2 n, m = 0;

	if( CHECKCONSTANT( cf, index, CONSTANT_Utf8 ) )
	{
		CONSTANT_Utf8_info* cu = (CONSTANT_Utf8_info*) cf->constant_pool[index];
		
		for( n = 0; n < cu->length; n++ )
		{
			if( cu->bytes[n] == '/' )
			{
				buf[m] = '.'; m++;
			}
			else if(( cu->bytes[n] >= 0x01 ) && ( cu->bytes[n] <= 0x7f ))
			{
				buf[m] = cu->bytes[n]; m++;
			}
		}
	}

	buf[m] = '\0';
	return m;
}

int sprintClassName( char* buf, ClassFile* cf, u2 index )
{
	if( CHECKCONSTANT( cf, index, CONSTANT_Class ) )
		return sprintSimpleName( buf, cf, ((CONSTANT_Class_info*) cf->constant_pool[index])->name_index );

	return 0;
}

#ifdef _DEBUG

void writeField( ClassFile* cf, field_info* field )
{
	char buf[255];

	if( !field )
		return;

	printf( "  " );

	if( field->access_flags & ACC_PUBLIC )
		printf( "public " );
	if( field->access_flags & ACC_PRIVATE )
		printf( "private " );
	if( field->access_flags & ACC_PROTECTED )
		printf( "protected " );
	if( field->access_flags & ACC_STATIC )
		printf( "static " );
	if( field->access_flags & ACC_FINAL )
		printf( "final " );
	if( field->access_flags & ACC_VOLATILE )
		printf( "volatile " );
	if( field->access_flags & ACC_TRANSIENT )
		printf( "transient " );

	sprintDescriptor( buf, cf, field->descriptor_index );
	printf( "%s ", buf );
	sprintSimpleName( buf, cf, field->name_index );
	printf( "%s;\n", buf );
}


int sprintNameAndType( char* buf, ClassFile* cf, u2 index )
{
	DECLARE_PRINTF();

	if( CHECKCONSTANT( cf, index, CONSTANT_NameAndType ) )
	{
		CONSTANT_NameAndType_info* natp = (CONSTANT_NameAndType_info*) cf->constant_pool[index];
		len = sprintMethodReturn( buf, cf, natp->descriptor_index ); buf += len; tlen += len;
		len = sprintSimpleName( buf, cf, natp->name_index ); buf += len; tlen += len;
		tlen += sprintMethodParameters( buf, cf, natp->descriptor_index );
	}

	RETURN_PRINTF();
}

#endif /* _DEBUG */

int sprintFieldRef( char* buf, ClassFile* cf, u2 index )
{
	DECLARE_PRINTF();

	if( CHECKCONSTANT( cf, index, CONSTANT_Fieldref ) )
	{
		CONSTANT_Fieldref_info* fp = (CONSTANT_Fieldref_info*) cf->constant_pool[index];
		len = sprintClassName( buf, cf, fp->class_index ); buf += len; tlen += len;
		PRINTF( ", " );
		if( CHECKCONSTANT( cf, fp->name_and_type_index, CONSTANT_NameAndType ) )
		{
			CONSTANT_NameAndType_info* natp = (CONSTANT_NameAndType_info*) cf->constant_pool[fp->name_and_type_index];
			len = sprintDescriptor( buf, cf, natp->descriptor_index ); buf += len; tlen += len;
			PRINTF( " " );
			tlen += sprintSimpleName( buf, cf, natp->name_index );
		}
	}

	RETURN_PRINTF();
}


int sprintMethodRef( char* buf, ClassFile* cf, u2 index )
{
	DECLARE_PRINTF();

	if( CHECKCONSTANT( cf, index, CONSTANT_Methodref ) )
	{
		CONSTANT_Methodref_info* mp = (CONSTANT_Methodref_info*) cf->constant_pool[index];
		len = sprintClassName( buf, cf, mp->class_index ); buf += len; tlen += len;
		PRINTF( ", " );
		if( CHECKCONSTANT( cf, mp->name_and_type_index, CONSTANT_NameAndType ) )
		{
			CONSTANT_NameAndType_info* natp = (CONSTANT_NameAndType_info*) cf->constant_pool[mp->name_and_type_index];
			len = sprintMethodReturn( buf, cf, natp->descriptor_index ); buf += len; tlen += len;
			len = sprintSimpleName( buf, cf, natp->name_index ); buf += len; tlen += len;
			tlen += sprintMethodParameters( buf, cf, natp->descriptor_index );
		}
	}

	RETURN_PRINTF();
}


int sprintInterfaceMethodRef( char* buf, ClassFile* cf, u2 index )
{
	DECLARE_PRINTF();

	if( CHECKCONSTANT( cf, index, CONSTANT_InterfaceMethodref ) )
	{
		CONSTANT_InterfaceMethodref_info* mp = (CONSTANT_InterfaceMethodref_info*) cf->constant_pool[index];
		len = sprintClassName( buf, cf, mp->class_index ); buf += len; tlen += len;
		PRINTF( ", " );
		if( CHECKCONSTANT( cf, mp->name_and_type_index, CONSTANT_NameAndType ) )
		{
			CONSTANT_NameAndType_info* natp = (CONSTANT_NameAndType_info*) cf->constant_pool[mp->name_and_type_index];
			len = sprintMethodReturn( buf, cf, natp->descriptor_index ); buf += len; tlen += len;
			len = sprintSimpleName( buf, cf, natp->name_index ); buf += len; tlen += len;
			tlen += sprintMethodParameters( buf, cf, natp->descriptor_index );
		}
	}

	RETURN_PRINTF();
}


int sprintMethodInfo( char* buf, method_info* mi )
{
	DECLARE_PRINTF();

	if( mi && mi->cf )
	{
		len = sprintClassName( buf, mi->cf, mi->cf->this_class ); buf += len; tlen += len;
		PRINTF( ", " );
		len = sprintMethodReturn( buf, mi->cf, mi->descriptor_index ); buf += len; tlen += len;
		len = sprintSimpleName( buf, mi->cf, mi->name_index ); buf += len; tlen += len;
		tlen += sprintMethodParameters( buf, mi->cf, mi->descriptor_index );
	}

	RETURN_PRINTF();
}

#ifdef _DEBUG

int sprintInstruction( char* buf, ClassFile* cf, u1* code, u4* n , int comments)
{
	int inst = code[*n];
	int pos = *n;

	DECLARE_PRINTF();

	PRINTF2( "%d: ", ((int) *n) );

	if( instructionTable[inst].size )
	{
		if( instructionTable[inst].name )
		{
			PRINTF( instructionTable[inst].name );
		}
		else
		{
			PRINTF2( "<unknown %d>", inst );
		}

		if( instructionTable[inst].format )
		{
			int comment = 0;
			int m;

			for( m = 0; instructionTable[inst].format[m]; m++ )
			{
				switch( instructionTable[inst].format[m] )
				{
				case 'i': case 't':
					comment = 1;
				case 'B':
					PRINTF2( " %d", code[pos+1] ); pos++;
					break;
				case 'c': case 'f': case 'l': case 'm': case 'v':
					comment = 1;
					PRINTF2( " #%d", PEEKU2(code,pos) ); pos+=2;
					break;
				case 'S':
					PRINTF2( " %d", (int16) PEEKU2(code,pos) ); pos+=2;
					break;
				case 'I':
					PRINTF2( " %d", ((int) PEEKU4(code,pos)) ); pos+=4;
					break;
				case 'J':
					PRINTF( " <HELP - long>" ); pos+=8;
					break;
				}
			}

			if( comment )
			{
				int m;

				PRINTF( "     /*" );
				pos = *n;

				for( m = 0; instructionTable[inst].format[m]; m++ )
				{
					switch( instructionTable[inst].format[m] )
					{
					case 't': PRINTF( " " ); len = sprintTypeName( buf, code[pos+1] ); buf += len; tlen += len; pos++; break;
					case 'i': PRINTF( " " ); len = sprintLiteral( buf, cf, code[pos+1] ); buf += len; tlen += len; pos++; break;
					case 'l': PRINTF( " " ); len = sprintLiteral( buf, cf, (u2) PEEKU2(code,pos) ); buf += len; tlen += len; pos+=2; break;
					case 'c': PRINTF( " " ); len = sprintClassName( buf, cf, (u2) PEEKU2(code,pos) ); buf += len; tlen += len; pos+=2; break;
					case 'f': PRINTF( " " ); len = sprintFieldRef( buf, cf, (u2) PEEKU2(code,pos) ); buf += len; tlen += len; pos+=2; break;
					case 'm': PRINTF( " " ); len = sprintMethodRef( buf, cf, (u2) PEEKU2(code,pos) ); buf += len; tlen += len; pos+=2; break;
					case 'v': PRINTF( " " ); len = sprintInterfaceMethodRef( buf, cf, (u2) PEEKU2(code,pos) ); buf += len; tlen += len; pos+=2; break;
					}
				}

				PRINTF( " */" );
			}
		}

		*n += (instructionTable[inst].size - 1);
	}
	else /* non-fixed size instruction or unknown instruction */
	{

		switch( inst )
		{
		case 0xaa:
			{
				int def, min, max, m;

				*n = ((*n)/4)*4 + 3;
				def = PEEKU4( code, *n ); (*n)+=4;
				min = PEEKU4( code, *n ); (*n)+=4;
				max = PEEKU4( code, *n ); (*n)+=4;
				PRINTF2( "tableswitch %d", min )
				PRINTF2( " to %d\n", max );

				for( m = 0; m < max - min + 1; m++ )
				{
					PRINTF2( "     %d", (min + m) );
					PRINTF2( ": %d\n", PEEKU4( code, *n ) );
					(*n)+=4;
				}

				PRINTF2( "     default: %d", def );
			}
			break;
		case 0xab:
			{
				int def, count, m;

				*n = ((*n)/4)*4 + 3;
				def = PEEKU4( code, *n ); (*n)+=4;
				count = PEEKU4( code, *n ); (*n)+=4;
				PRINTF2( "lookupswitch %d\n", count );

				for( m = 0; m < count; m++ )
				{
					PRINTF2( "     %d", PEEKU4( code, *n ) ); (*n)+=4;
					PRINTF2( ": %d\n", PEEKU4( code, *n ) ); (*n)+=4;
				}

				PRINTF2( "     default: %d", def );
			}
			break;
		default:
			PRINTF2( "<unknown %d>", inst );
			break;
		}
	}

	RETURN_PRINTF();
}


void writeAttribute( ClassFile* cf, attribute_info* attr, const char* indent )
{
	char buf[255];

	printf( indent );
	sprintSimpleName( buf, cf, attr->attribute_name_index );
	printf( "%s (length: %d)\n", buf, ((int) attr->attribute_length) );

	if( checkSimpleName( cf, attr->attribute_name_index, "Code" ) )
	{
		Code_attribute* ca = (Code_attribute*) attr;
		u1* code;
		u4 n, max;
		
		printf( "%smax stack  : %d\n", indent, ca->max_stack );
		printf( "%smax locals : %d\n", indent, ca->max_locals );
		printf( "%scode length: %d\n", indent, ((int) ca->code_length) );
		
		code = (u1*) ca->code;
		max = ca->code_length;
		
		for( n = 0; n < max; n++ )
		{
			printf( indent );
			sprintInstruction( buf, cf, code, &n, 1 );
			printf( "%s\n", buf );
		}
	}
	else if( checkSimpleName( cf, attr->attribute_name_index, "SourceFile" ) )
	{
		SourceFile_attribute* sfa = (SourceFile_attribute*) attr;
		sprintSimpleName( buf, cf, sfa->sourcefile_index );
		printf( "source file: %d     /* %s */\n", getU2( sfa->sourcefile_index ), buf );
	}
}


void writeMethod( ClassFile* cf, method_info* method, int level )
{
	char buf[255];
	int show = 1;
	u2 n;

	if( !method )
		return;

	printf( "  " );

	if( method->access_flags & ACC_PUBLIC )
		printf( "public " );
	if( method->access_flags & ACC_PRIVATE )
		printf( "private " );
	if( method->access_flags & ACC_PROTECTED )
		printf( "protected " );
	if( method->access_flags & ACC_STATIC )
		printf( "static " );
	if( method->access_flags & ACC_FINAL )
		printf( "final " );
	if( method->access_flags & ACC_SYNCHRONIZED )
		printf( "synchronized " );
	if( method->access_flags & ACC_NATIVE )
		printf( "native " );
	if( method->access_flags & ACC_ABSTRACT )
		printf( "abstract " );

	sprintMethodReturn( buf, cf, method->descriptor_index ); printf( "%s", buf );
	sprintSimpleName( buf, cf, method->name_index ); printf( "%s", buf );
	sprintMethodParameters( buf, cf, method->descriptor_index ); printf( "%s", buf );
	
	for( n = 0; n < method->attributes_count; n++ )
	{
		if( checkSimpleName( cf, method->attributes[n]->attribute_name_index, "Code" ) )
		{
			if( level & 0x01 )
			{
				if( show ) { printf( "\n  {\n" ); show = 0; }
				writeAttribute( cf, method->attributes[n], "    " );
			}
		}
		else if( level & 0x08 )
		{
			if( show ) { printf( "\n  {\n" ); show = 0; }
			writeAttribute( cf, method->attributes[n], "    " );
		}
	}
	
	if( show )
		printf( ";\n" );
	else
		printf( "  }\n" );
}


void disassemble( ClassFile *cf, int level )
{
	char buf[255];
	cp_info *cp;
	u2 n;

	printf( "magic number: 0x%08x\n", cf->magic );

	if( cf->magic != 0xcafebabe )
		return;

	if( (cf->major_version == 45) && (cf->minor_version == 3) )
		printf( "version     : %d.%d (JDK 1.0.2)\n", cf->major_version, cf->minor_version );
	else
		printf( "version     : %d.%d\n", cf->major_version, cf->minor_version );

	printf( "constants   : %d\n", cf->constant_pool_count );

	if( level & 0x02 )
	{
		for( n = 0; n < cf->constant_pool_count; n++ )
		{
			cp = cf->constant_pool[n];
			
			if( !cp )
				continue;
			
			printf( "  %d", n );
			
			switch( cp->tag )
			{
			case CONSTANT_Utf8:
				printf( "  string \"" ); writeUTF8( ((CONSTANT_Utf8_info*) cp)->bytes, ((CONSTANT_Utf8_info*) cp)->length ); printf( "\"\n" );
				break;
			case CONSTANT_Integer:
				printf( "  int %d\n", ((int) ((CONSTANT_Integer_info*) cp)->bytes) );
				break;
			case CONSTANT_Float:
				printf( "  float \n" );
				break;
			case CONSTANT_Long:
				printf( "  long \n" );
				break;
			case CONSTANT_Double:
				printf( "  double \n" );
				break;
			case CONSTANT_Class:
				sprintSimpleName( buf, cf, ((CONSTANT_Class_info*) cp)->name_index );
				printf( "  class %d     /* %s */\n", ((CONSTANT_Class_info*) cp)->name_index, buf );
				break;
			case CONSTANT_String:
				printf( "  java.lang.String %d     /* \"", ((CONSTANT_String_info*) cp)->string_index );
				if( CHECKCONSTANT( cf, ((CONSTANT_String_info*) cp)->string_index, CONSTANT_Utf8 ) )
				{
					CONSTANT_Utf8_info* utfp = (CONSTANT_Utf8_info*) cf->constant_pool[ ((CONSTANT_String_info*) cp)->string_index ];
					writeUTF8( utfp->bytes, utfp->length );
				}
				printf( "\" */\n" );
				break;
			case CONSTANT_Fieldref:
				sprintFieldRef( buf, cf, n );
				printf( "  FieldRef %d %d     /* %s */\n",
					((CONSTANT_Fieldref_info*) cp)->class_index,
					((CONSTANT_Fieldref_info*) cp)->name_and_type_index,
					buf );
				break;
			case CONSTANT_Methodref:
				sprintMethodRef( buf, cf, n );
				printf( "  MethodRef %d %d     /* %s */\n",
					((CONSTANT_Methodref_info*) cp)->class_index,
					((CONSTANT_Methodref_info*) cp)->name_and_type_index,
					buf );
				break;
			case CONSTANT_InterfaceMethodref:
				printf( "  InterfaceMethodRef %d %d\n", ((CONSTANT_InterfaceMethodref_info*) cp)->class_index, ((CONSTANT_InterfaceMethodref_info*) cp)->name_and_type_index );
				break;
			case CONSTANT_NameAndType:
				sprintNameAndType( buf, cf, n );
				printf( "  NameAndType %d %d     /* %s */\n", ((CONSTANT_NameAndType_info*) cp)->name_index, ((CONSTANT_NameAndType_info*) cp)->descriptor_index, buf );
				break;
			}
		}
	}

	printf( "access flags: %d\n", cf->access_flags );
	printf( "this class  : %d\n", cf->this_class );
	printf( "super class : %d\n", cf->super_class );
	printf( "interfaces  : %d\n", cf->interfaces_count );
	printf( "fields      : %d\n", cf->fields_count );
	printf( "methods     : %d\n", cf->methods_count );
	printf( "attributes  : %d\n", cf->attributes_count );

	if( (level & 0x04) && cf->attributes_count )
	{
		printf( "\n" );
		for( n = 0; (n < cf->attributes_count) && cf->attributes[n]; n++ )
			writeAttribute( cf, cf->attributes[n], "" );
	}

	printf( "\n" );

	if( cf->access_flags & ACC_PUBLIC )
		printf( "public " );

	if( cf->access_flags & ACC_FINAL )
		printf( "final " );

	if( cf->access_flags & ACC_SUPER )
		printf( "super " );

	if( cf->access_flags & ACC_INTERFACE )
		printf( "interface " );
	else
		printf( "class " );

	if( cf->access_flags & ACC_ABSTRACT )
		printf( "abstract " );

	sprintClassName( buf, cf, cf->this_class );
	printf( "%s\n", buf );

	if( cf->super_class )
	{
		sprintClassName( buf, cf, cf->super_class );
		printf( "  extends %s\n", buf );
	}

	if( cf->interfaces_count )
	{
		printf( "  implements ");
		printf( "\n" );
	}

	printf( "{\n" );

	for( n = 0; (n < cf->fields_count) && cf->fields[n]; n++ )
		writeField( cf, cf->fields[n] );

	for( n = 0; (n < cf->methods_count) && cf->methods[n]; n++ )
		writeMethod( cf, cf->methods[n], level );

	printf( "}\n" );
}

#endif /* _DEBUG */
