/******************************************************************************
 *
 * Filename: integer.c
 *
 * Author:   Robert Thornburrow
 * Project:  JavaVM
 * Created:  4th March 2002
 *
 * Purpose:  Include file for integer instructions.
 *
 * (c) Copyright R.H.Thornburrow, 2002
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *****************************************************************************/

void i2b( Frame* fp )
{
	int32 value, result;

	INSTRUCTION( "%p: i2b\n", fp->pc );

	POP_INT( fp, value );
	result = (int8) value;

	if( value < 0 )
		result |= 0xffffff00;

	PUSH_INT( fp, result );

	if( arithmetic_debug & 2 )
		printf( "%p: i2b: 0x%08x => 0x%02x\n", fp->pc, value, (u1) result );
}


void i2c( Frame* fp )
{
	u4 value = 0;
	int32 i;
	u1* c = (u1*) &value;

	INSTRUCTION( "%p: i2c\n", fp->pc );

	POP_INT( fp, i );

	if( i > 0xff )
		printf( "WARNING: i2c (0x%04x)\n", (u2) i );

	*c = (u1) i;
	PUSH_INT( fp, value );

	if( arithmetic_debug & 2 )
		printf( "%p: i2c: 0x%08x => 0x%04x\n", fp->pc, i, (u2) value );
}


void i2d( Frame* fp )
{
	u4 ivalue;
	u4 dvalue[2] = { 0, 0 };
	int32* i = (int32*) &ivalue;
	double* d = (double*) dvalue;

	INSTRUCTION( "%p: i2d\n", fp->pc );

	POP_INT( fp, ivalue );
	*d = (double) *i;

	PUSH_DOUBLE( fp, dvalue );

	ARITHMETIC_DEBUG( printf( "%p: i2d: 0x%08x => %f\n", fp->pc, *i, *d ); )
}


void i2f( Frame* fp )
{
	u4 value = 0;
	int32 i;
	float* f = (float*) &value;

	INSTRUCTION( "%p: i2f\n", fp->pc );

	POP_INT( fp, i );
	*f = (float) i;
	PUSH_FLOAT( fp, value );

	ARITHMETIC_DEBUG( printf( "%p: i2f: 0x%08x => %f\n", fp->pc, i, *f ); )
}


void i2s( Frame* fp )
{
	int32 value, result;

	INSTRUCTION( "%p: i2s\n", fp->pc );

	POP_INT( fp, value );
	result = (int16) value;

	if( value < 0 )
		result |= 0xffff0000;

	PUSH_INT( fp, result );

	if( arithmetic_debug & 2 )
		printf( "%p: i2s: 0x%08x => 0x%04x\n", fp->pc, value, (int16) result );
}


void i2l( Frame* fp )
{
	u4 resultw[2] = { 0, 0 };
	int32 value;
	int64* lr = (int64*) resultw;

	INSTRUCTION( "%p: i2l\n", fp->pc );

	POP_INT( fp, value );
	*lr = (int64) value;
	PUSH_LONG( fp, resultw );

	if( arithmetic_debug  & 2)
		printf( "%p: i2l: 0x%08x => 0x%08x%08x\n", fp->pc, value, resultw[1], resultw[0] );
}


void iadd( Frame* fp )
{
	int32 value1, value2, result;

	INSTRUCTION( "%p: iadd\n", fp->pc );

	POP_INT( fp, value2 );
	POP_INT( fp, value1 );
	result = value1 + value2;
	PUSH_INT( fp, result );

	if( arithmetic_debug & 2 )
		printf( "%p: iadd: %d + %d => %d\n", fp->pc, (int) value1, (int) value2, (int) result );
}


int iaload( Stack* stp )
{
	Frame *fp = stp->fp;
	int32 index;
	ObjRef* array;
	u4* data;

	INSTRUCTION( "%p: iaload\n", fp->pc );

	POP_INT( fp, index );
	POP_OBJREF( fp, array );

	if( !array )
		return newCompleteException( stp, NullPointerException, "iaload" );

	if( (index < 0) || (index >= (int32) array->length) )
	{
		char buf[32];
		sprintf( buf, "%d", (int) index );
		return newCompleteException( stp, ArrayIndexOutOfBoundsException, buf );
	}

	if( (array->cf != NULL) || (array->type != T_INT) || (array->data == NULL) )
		return newCompleteException( stp, InternalError, "Bad int[] object reference" );

	data = (u4*) array->data;

	PUSH_INT( fp, data[index] );

	if( arithmetic_debug )
		printf( "%p: iaload: 0x%p [ %d ] => %d\n", fp->pc, data, (int) index, (int) data[index] );

	return 0;
}


void iand( Frame* fp )
{
	int32 value1, value2, result;

	INSTRUCTION( "%p: iand\n", fp->pc );

	POP_INT( fp, value2 );
	POP_INT( fp, value1 );
	result = value1 & value2;
	PUSH_INT( fp, result );

	if( arithmetic_debug & 2 )
		printf( "%p: iand: 0x%08x & 0x%08x => 0x%08x\n", fp->pc, value1, value2, result );
}


int iastore( Stack* stp )
{
	Frame *fp = stp->fp;
	u4 value;
	int32 index;
	ObjRef* array;
	u4* data;

	INSTRUCTION( "%p: iastore\n", fp->pc );

	POP_INT( fp, value );
	POP_INT( fp, index );
	POP_OBJREF( fp, array );

	if( !array )
		return newCompleteException( stp, NullPointerException, "iastore" );

	if( (index < 0) || (index >= (int32) array->length) )
	{
		char buf[32];
		sprintf( buf, "%d", (int) index );
		return newCompleteException( stp, ArrayIndexOutOfBoundsException, buf );
	}

	if( (array->cf != NULL) || (array->type != T_INT) || (array->data == NULL) )
		return newCompleteException( stp, InternalError, "Bad int[] object reference" );

	data = (u4*) array->data;

	data[index] = (u4) value;

	ARITHMETIC_DEBUG( printf( "%p: iastore: 0x%p [ %d ] <= %d\n", fp->pc, data, (int) index, (int) value ); )
	return 0;
}


int idiv( Stack* stp )
{
	Frame *fp = stp->fp;
	int32 value1, value2, result;

	INSTRUCTION( "%p: idiv\n", fp->pc );

	POP_INT( fp, value2 );
	POP_INT( fp, value1 );

	if( value2 == 0 )
		return newCompleteException( stp, ArithmeticException, "divide by zero" );

	/* special case - see idiv documentation */
	if( (value1 == MIN_INTEGER ) && (value2 == -1) )
	{
		PUSH_INT( fp, value1 );
		return 0;
	}

	result = value1 / value2;
	PUSH_INT( fp, result );

	if( arithmetic_debug & 2 )
		printf( "%p: idiv: %d / %d => %d\n", fp->pc, (int) value1, (int) value2, (int) result );

	return 0;
}


void iconst_n( Frame *fp, int32 n )
{
	INSTRUCTION( "%p: iconst_n %d\n", fp->pc, (int) n );
	PUSH_INT( fp, n );
}


void iinc( Frame* fp, u1 index, u1 increment )
{
	int32 value;
    int32 inc = (increment > 127) ? (int32) (increment - 256) : (int32) increment;

	INSTRUCTION( "%p: iinc %d %d\n", fp->pc, index, increment );

	GET_LINT( fp, index, value );
	value += inc;
	PUT_LINT( fp, index, value );
}


void iinc_w( Frame* fp, u2 index, u2 increment )
{
	int32 value;
    int32 inc = (increment > 32767) ? (int32) (increment - 65536) : (int32) increment;

	INSTRUCTION( "%p: iinc_w %d %d\n", fp->pc, index, increment );

	GET_LINT( fp, index, value );
	value += inc;
	PUT_LINT( fp, index, value );
}


void iload_n( Frame* fp, u2 index )
{
	INSTRUCTION( "%p: iload_%d\n", fp->pc, index );
	PUSH_LINT( fp, index );
}


void imul( Frame* fp )
{
	int32 value1, value2, result;

	INSTRUCTION( "%p: imul\n", fp->pc );

	POP_INT( fp, value2 );
	POP_INT( fp, value1 );
	result = value1 * value2;
	PUSH_INT( fp, result );

	if( arithmetic_debug & 2 )
		printf( "%p: imul: %d * %d => %d\n", fp->pc, (int) value1, (int) value2, (int) result );
}


void ineg( Frame* fp )
{
	int32 value, result;

	INSTRUCTION( "%p: ineg\n", fp->pc );

	POP_INT( fp, value );
	result = - value;
	PUSH_INT( fp, result );

	if( arithmetic_debug & 2 )
		printf( "%p: ineg: %d => %d\n", fp->pc, (int) value, (int) result );
}


void ior( Frame* fp )
{
	int32 value1, value2, result;

	INSTRUCTION( "%p: ior\n", fp->pc );

	POP_INT( fp, value2 );
	POP_INT( fp, value1 );
	result = value1 | value2;
	PUSH_INT( fp, result );

	if( arithmetic_debug & 2 )
		printf( "%p: ior: 0x%08x | 0x%08x => 0x%08x\n", fp->pc, value1, value2, result );
}


int irem( Stack* stp )
{
	Frame *fp = stp->fp;
	int32 value1, value2, result;

	INSTRUCTION( "%p: irem\n", fp->pc );

	POP_INT( fp, value2 );
	POP_INT( fp, value1 );

	if( value2 == 0 )
		return newCompleteException( stp, ArithmeticException, "divide by zero" );

	result = value1 % value2;
	PUSH_INT( fp, result );

	if( arithmetic_debug & 2 )
		printf( "%p: irem: %d %% %d => %d\n", fp->pc, (int) value1, (int) value2, (int) result );

	return 0;
}


int ireturn( Stack* stp )
{
	Frame* fp = stp->fp;
	int32 value;
	int stopOnPop = stp->fp->stopOnPop;

	INSTRUCTION( "%p: ireturn\n", fp->pc );

	POP_INT( fp, value );
	stp->fp = fp->next;
	freeStackFrame( fp, fp->mi );
	fp = stp->fp;

	if( fp )
	{
		fp->prev = NULL;
		PUSH_INT( fp, value );
	}

#ifdef _DEBUG
	if( fp && method_debug )
	{
		char buf[255];
		sprintMethodInfo( buf, fp->mi );
		printf( "Returned to method %s\n", buf );
	}
#endif /* _DEBUG */

	if( callstack )
		(*callstack)( fp, 0 );

	return ( (!fp || stopOnPop) ? 1 : 0 );
}


void ishl( Frame* fp )
{
	int32 value1, value2, result;

	INSTRUCTION( "%p: ishl\n", fp->pc );

	POP_INT( fp, value2 );

	if( value2 > 0x1f )
		printf( "WARNING: ishl being asked to shift by 0x%08x\n", value2 );

	POP_INT( fp, value1 );
	result = value1 << (value2 & 0x1f);
	PUSH_INT( fp, result );

	if( arithmetic_debug & 2 )
		printf( "%p: ishl: 0x%08x << 0x%08x => 0x%08x\n", fp->pc, value1, value2, result );
}


void ishr( Frame* fp )
{
	int32 value1, value2, result;

	INSTRUCTION( "%p: ishr\n", fp->pc );

	POP_INT( fp, value2 );

	if( value2 > 0x1f )
		printf( "WARNING: ishr being asked to shift by 0x%08x\n", value2 );

	POP_INT( fp, value1 );
	result = value1 >> (value2 & 0x1f);
	PUSH_INT( fp, result );

	if( arithmetic_debug & 2 )
		printf( "%p: ishr: 0x%08x >> 0x%08x => 0x%08x\n", fp->pc, value1, value2, result );
}


void istore_n( Frame* fp, u2 index )
{
	INSTRUCTION( "%p: istore_%d\n", fp->pc, index );
	POP_LINT( fp, index );
}


void isub( Frame* fp )
{
	int32 value1, value2, result;

	INSTRUCTION( "%p: isub\n", fp->pc );

	POP_INT( fp, value2 );
	POP_INT( fp, value1 );
	result = value1 - value2;
	PUSH_INT( fp, result );

	if( arithmetic_debug & 2 )
		printf( "%p: isub: %d - %d => %d\n", fp->pc, (int) value1, (int) value2, (int) result );
}


void iushr( Frame* fp )
{
	int32 value1, value2, result;

	INSTRUCTION( "%p: iushr\n", fp->pc );

	POP_INT( fp, value2 );

	if( value2 > 0x1f )
		printf( "WARNING: iushr being asked to shift by 0x%08x\n", value2 );

	POP_INT( fp, value1 );

	if( value1 >= 0 )
		result = value1 >> (value2 & 0x1f);
	else
		result = (value1 >> (value2 & 0x1f)) + (2 << ~(value2 & 0x1f));

	PUSH_INT( fp, result );

	ARITHMETIC_DEBUG( printf( "%p: iushr: 0x%08x >> 0x%08x => 0x%08x\n", fp->pc, value1, value2, result ); )
}


void ixor( Frame* fp )
{
	int32 value1, value2, result;

	INSTRUCTION( "%p: ixor\n", fp->pc );

	POP_INT( fp, value2 );
	POP_INT( fp, value1 );
	result = value1 ^ value2;
	PUSH_INT( fp, result );

	ARITHMETIC_DEBUG( printf( "%p: ixor: 0x%08x ^ 0x%08x => 0x%08x\n", fp->pc, value1, value2, result ); )
}


