/******************************************************************************
 *
 * Filename: native.c
 *
 * Author:   Robert Thornburrow
 * Project:  JavaVM
 * Created:  19th January 1998
 *
 * Purpose:  Native method implementations for command line runtime.
 *
 * (c) Copyright R.H.Thornburrow, 1998
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *****************************************************************************/

#include <stdio.h>

#include "config.h"
#include "types.h"
#include "disasm.h"
#include "util.h"

#include "../native/java_io_File.h"
#include "../native/java_io_FileDescriptor.h"
#include "../native/java_io_FileInputStream.h"
#include "../native/java_io_FileOutputStream.h"
#include "../native/java_io_ObjectInputStream.h"
#include "../native/java_io_ObjectOutputStream.h"
#include "../native/java_io_ObjectStreamClass.h"
#include "../native/java_lang_Class.h"
#include "../native/java_lang_ClassLoader.h"
#include "../native/java_lang_Math.h"
#include "../native/java_lang_Object.h"
#include "../native/java_lang_Runtime.h"
#include "../native/java_lang_SecurityManager.h"
#include "../native/java_lang_System.h"
#include "../native/java_lang_Thread.h"
#include "../native/java_lang_Throwable.h"

#include "native.h"


FILE* fdList[255];
int fdLast = 2;


#define STRING_MAX 256 /* 1024 */

#define CHECK_OBJREF_NOT_NULL( objref, stp )  	\
	{ if( !objref ) return newCompleteException( stp, NullPointerException, "" ); }

#define CHECK_OBJREF_CF_AND_DATA( objref, stp )  \
	{ if( !objref->cf || (objref->type != 0) ) return newCompleteException( stp, InternalError, "corrupt object reference" ); }

#define CHECK_OBJREF_TYPE( objref, type, stp )	\
	{ if( !checkType(objref->cf,type) ) return newCompleteException( stp, InternalError, "object reference not of type \"" type "\"" ); }

#define CHECK_OBJREF_DATA( objref, stp ) \
	{ if( !objref->data ) return newCompleteException( stp, InternalError, "bad object reference" ); }

#define CHECK_OBJREF( objref, fp )  { CHECK_OBJREF_NOT_NULL(objref,fp); CHECK_OBJREF_CF_AND_DATA(objref,fp); }

#define CHECK_OBJREF_AND_TYPE( objref, type, fp )  { \
	CHECK_OBJREF_NOT_NULL(objref,fp); CHECK_OBJREF_CF_AND_DATA(objref,fp); \
	CHECK_OBJREF_TYPE(objref,type,fp); CHECK_OBJREF_DATA(objref,fp); }


/******************************************************************************
 *
 * Wrapper functions
 *
 *****************************************************************************/

typedef void     (*fn_V)( JNIEnv* env, jobject self );
typedef jboolean (*fn_Z)( JNIEnv* env, jobject self );
typedef jint     (*fn_I)( JNIEnv* env, jobject self );
typedef jlong    (*fn_J)( JNIEnv* env, jobject self );
typedef jobject  (*fn_L)( JNIEnv* env, jobject self );


/* void fn() - ()V */
int wrapper_V( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_V fn = (fn_V) gfn;
	Frame* fp = stp->fp;
	ObjRef* self;

	POP_OBJREF( fp, self );
	CHECK_OBJREF( self, stp );

	(*fn)( stp, self );

	if( stp->ex || fp->ex )
		return stp->retcode;

	return 0;
}


/* boolean fn() - ()Z */
int wrapper_Z( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_Z fn = (fn_Z) gfn;
	Frame* fp = stp->fp;
	ObjRef* self;
	int32 ret;

	POP_OBJREF( fp, self );
	CHECK_OBJREF( self, stp );

	ret = (*fn)( stp, self );

	if( stp->ex || fp->ex )
		return stp->retcode;

	PUSH_INT( fp, ret );
	return 0;
}


/* int fn() - ()I */
int wrapper_I( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_I fn = (fn_I) gfn;
	Frame* fp = stp->fp;
	ObjRef* self;
	int32 ret;

	POP_OBJREF( fp, self );
	CHECK_OBJREF( self, stp );

	ret = (*fn)( stp, self );

	if( stp->ex || fp->ex )
		return stp->retcode;

	PUSH_INT( fp, ret );
	return 0;
}


/* int fn() - ()J */
int wrapper_J( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_J fn = (fn_J) gfn;
	Frame* fp = stp->fp;
	ObjRef* self;
	u4 retw[2];
	int64* ret = (int64*) retw;

	POP_OBJREF( fp, self );
	CHECK_OBJREF( self, stp );

	*ret = (*fn)( stp, self );

	if( stp->ex || fp->ex )
		return stp->retcode;

	PUSH_LONG( fp, retw );
	return 0;
}


/* object fn() - ()Lxxx; */
int wrapper_L( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_L fn = (fn_L) gfn;
	Frame* fp = stp->fp;
	ObjRef *self, *ret;

	POP_OBJREF( fp, self );
	CHECK_OBJREF( self, stp );

	ret = (*fn)( stp, self );

	if( stp->ex || fp->ex )
		return stp->retcode;

	PUSH_OBJREF( fp, ret );
	return 0;
}


/*****************************************************************************/


typedef void     (*fn_ZV)( JNIEnv* env, jobject self, jboolean arg1 );
typedef void     (*fn_IV)( JNIEnv* env, jobject self, jint arg1 );
typedef void     (*fn_JV)( JNIEnv* env, jobject self, jlong arg1 );
typedef void     (*fn_LV)( JNIEnv* env, jobject self, jobject arg1 );

typedef jboolean (*fn_LZ)( JNIEnv* env, jobject self, jobject arg1 );

typedef jint     (*fn_LI)( JNIEnv* env, jobject self, jobject arg1 );

typedef jlong    (*fn_JJ)( JNIEnv* env, jobject self, jlong arg1 );

typedef jobject  (*fn_LL)( JNIEnv* env, jobject self, jobject arg1 );


/* void fn(boolean) - (Z)V */
int wrapper_ZV( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_ZV fn = (fn_ZV) gfn;
	Frame* fp = stp->fp;
	ObjRef* self;
	int arg1;
	
	POP_INT( fp, arg1 );
	POP_OBJREF( fp, self );
	CHECK_OBJREF( self, stp );

	(*fn)( stp, self, arg1 );

	if( stp->ex || fp->ex )
		return stp->retcode;

	return 0;
}


/* void fn(int) - (I)V */
int wrapper_IV( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_IV fn = (fn_IV) gfn;
	Frame* fp = stp->fp;
	ObjRef* self;
	int32 arg1;
	
	POP_INT( fp, arg1 );
	POP_OBJREF( fp, self );
	CHECK_OBJREF( self, stp );

	(*fn)( stp, self, arg1 );

	if( stp->ex || fp->ex )
		return stp->retcode;

	return 0;
}


/* void fn(long) - (J)V */
int wrapper_JV( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_JV fn = (fn_JV) gfn;
	Frame* fp = stp->fp;
	ObjRef* self;
	u4 value[2];
	int64* arg1 = (int64*) value;

	POP_LONG( fp, value );
	POP_OBJREF( fp, self );
	CHECK_OBJREF( self, stp );

	(*fn)( stp, self, *arg1 );

	if( stp->ex || fp->ex )
		return stp->retcode;

	return 0;
}


/* void fn(object) - (Lxxx;)V */
int wrapper_LV( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_LV fn = (fn_LV) gfn;
	Frame* fp = stp->fp;
	ObjRef *self, *arg1;
	
	POP_OBJREF( fp, arg1 );
	POP_OBJREF( fp, self );
	CHECK_OBJREF( self, stp );
	CHECK_OBJREF( arg1, stp );

	(*fn)( stp, self, arg1 );

	if( stp->ex || fp->ex )
		return stp->retcode;

	return 0;
}


/* boolean fn(object) - (Lxxx;)Z */
int wrapper_LZ( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_LZ fn = (fn_LZ) gfn;
	Frame* fp = stp->fp;
	ObjRef *self, *arg1;
	int ret;
	
	POP_OBJREF( fp, arg1 );
	POP_OBJREF( fp, self );
	CHECK_OBJREF( self, stp );

	ret = (*fn)( stp, self, arg1 );

	if( stp->ex || fp->ex )
		return stp->retcode;

	PUSH_INT( fp, ret );
	return 0;
}


/* int fn(object) - (Lxxx;)I */
int wrapper_LI( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_LI fn = (fn_LI) gfn;
	Frame* fp = stp->fp;
	ObjRef *self, *arg1;
	int32 ret;
	
	POP_OBJREF( fp, arg1 );
	POP_OBJREF( fp, self );
	CHECK_OBJREF( self, stp );

	ret = (*fn)( stp, self, arg1 );

	if( stp->ex || fp->ex )
		return stp->retcode;

	PUSH_INT( fp, ret );
	return 0;
}


/* long fn(long) - (J)J */
int wrapper_JJ( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_JJ fn = (fn_JJ) gfn;
	Frame* fp = stp->fp;
	ObjRef* self;
	u4 value[2];
	int64* l = (int64*) value;
	int64 arg1, ret;

	POP_LONG( fp, value );
	arg1 = *l;
	POP_OBJREF( fp, self );
	CHECK_OBJREF( self, stp );

	ret = (*fn)( stp, self, arg1 );

	if( stp->ex || fp->ex )
		return stp->retcode;

	*l = ret;
	PUSH_LONG( fp, value );
	return 0;
}


/* object fn(object) - (Lxxx;)Lxxx; */
int wrapper_LL( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_LL fn = (fn_LL) gfn;
	Frame* fp = stp->fp;
	ObjRef *self, *arg1, *ret;
	
	POP_OBJREF( fp, arg1 );
	POP_OBJREF( fp, self );
	CHECK_OBJREF( self, stp );

	ret = (*fn)( stp, self, arg1 );

	if( stp->ex || fp->ex )
		return stp->retcode;

	PUSH_OBJREF( fp, ret );
	return 0;
}


/*****************************************************************************/


typedef jobject  (*fn_LLL)( JNIEnv* env, jobject self, jobject arg1, jobject arg2 );
typedef void     (*fn_LLLV)( JNIEnv* env, jobject self, jobject arg1, jobject arg2, jobject arg3 );
typedef void     (*fn_LIIV)( JNIEnv* env, jobject self, jobject arg1, jint arg2, jint arg3 );
typedef jint     (*fn_LIII)( JNIEnv* env, jobject self, jobject arg1, jint arg2, jint arg3 );


/* object fn(object,object) - (Lxxx;Lxxx;)Lxxx; */
int wrapper_LLL( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_LLL fn = (fn_LLL) gfn;
	Frame* fp = stp->fp;
	ObjRef *self, *arg1, *arg2, *ret;
	
	POP_OBJREF( fp, arg2 );
	POP_OBJREF( fp, arg1 );
	POP_OBJREF( fp, self );
	CHECK_OBJREF( self, stp );

	ret = (*fn)( stp, self, arg1, arg2 );

	if( stp->ex || fp->ex )
		return stp->retcode;

	PUSH_OBJREF( fp, ret );
	return 0;
}


/* void fn(object,object,object) - (Lxxx;Lxxx;Lxxx;)Lxxx; */
int wrapper_LLLV( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_LLLV fn = (fn_LLLV) gfn;
	Frame* fp = stp->fp;
	ObjRef *self, *arg1, *arg2, *arg3;
	
	POP_OBJREF( fp, arg3 );
	POP_OBJREF( fp, arg2 );
	POP_OBJREF( fp, arg1 );
	POP_OBJREF( fp, self );
	CHECK_OBJREF( self, stp );

	(*fn)( stp, self, arg1, arg2, arg3 );

	if( stp->ex || fp->ex )
		return stp->retcode;

	return 0;
}


/* int fn(object,int,int) - (Lxxx;II)I */
int wrapper_LIIV( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_LIII fn = (fn_LIII) gfn;
	Frame* fp = stp->fp;
	ObjRef *self, *arg1;
	int32 arg2, arg3;
	
	POP_INT( fp, arg3 );
	POP_INT( fp, arg2 );
	POP_OBJREF( fp, arg1 );
	POP_OBJREF( fp, self );
	CHECK_OBJREF( self, stp );

	(*fn)( stp, self, arg1, arg2, arg3 );

	if( stp->ex || fp->ex )
		return stp->retcode;

	return 0;
}


/* int fn(object,int,int) - (Lxxx;II)I */
int wrapper_LIII( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_LIII fn = (fn_LIII) gfn;
	Frame* fp = stp->fp;
	ObjRef *self, *arg1;
	int32 arg2, arg3, ret;
	
	POP_INT( fp, arg3 );
	POP_INT( fp, arg2 );
	POP_OBJREF( fp, arg1 );
	POP_OBJREF( fp, self );
	CHECK_OBJREF( self, stp );

	ret = (*fn)( stp, self, arg1, arg2, arg3 );

	if( stp->ex || fp->ex )
		return stp->retcode;

	PUSH_INT( fp, ret );
	return 0;
}


/*****************************************************************************/


typedef jlong    (*fn_static_J)( JNIEnv* env, jclass staticself );
typedef jobject  (*fn_static_L)( JNIEnv* env, jclass staticself );


/* static int fn() - ()J */
int wrapper_static_J( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_static_J fn = (fn_static_J) gfn;
	Frame* fp = stp->fp;
	u4 retw[2];
	int64* ret = (int64*) retw;

	*ret = (*fn)( stp, staticself );

	if( stp->ex || fp->ex )
		return stp->retcode;

	PUSH_LONG( fp, retw );
	return 0;
}


/* static object fn() - ()L */
int wrapper_static_L( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_static_L fn = (fn_static_L) gfn;
	Frame* fp = stp->fp;
	ObjRef* ret;

	ret = (*fn)( stp, staticself );

	if( stp->ex || fp->ex )
		return stp->retcode;

	PUSH_OBJREF( fp, ret );
	return 0;
}


/*****************************************************************************/


typedef void     (*fn_static_IV)( JNIEnv* env, jclass staticself, jint arg1 );
typedef void     (*fn_static_LV)( JNIEnv* env, jclass staticself, jobject arg1 );


/* static void fn(int) - (I)V */
int wrapper_static_IV( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_static_IV fn = (fn_static_IV) gfn;
	Frame* fp = stp->fp;
	int32 arg1;
	
	POP_INT( fp, arg1 );

	(*fn)( stp, staticself, arg1 );

	if( stp->ex || fp->ex )
		return stp->retcode;

	return 0;
}


/* static void fn(object) - (Lxxx;)V */
int wrapper_static_LV( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_static_LV fn = (fn_static_LV) gfn;
	Frame* fp = stp->fp;
	ObjRef* arg1;
	
	POP_OBJREF( fp, arg1 );

	(*fn)( stp, staticself, arg1 );

	if( stp->ex || fp->ex )
		return stp->retcode;

	return 0;
}


/*****************************************************************************/


typedef jboolean (*fn_static_LZ)( JNIEnv* env, jclass staticself, jobject arg1 );
typedef jint     (*fn_static_LI)( JNIEnv* env, jclass staticself, jobject arg1 );
typedef jlong    (*fn_static_LJ)( JNIEnv* env, jclass staticself, jobject arg1 );
typedef jdouble  (*fn_static_DD)( JNIEnv* env, jclass staticself, jdouble arg1 );
typedef jobject  (*fn_static_LL)( JNIEnv* env, jclass staticself, jobject arg1 );


/* static boolean fn(object) - (Lxxx;)Z */
int wrapper_static_LZ( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_static_LZ fn = (fn_static_LZ) gfn;
	Frame* fp = stp->fp;
	ObjRef* arg1;
	int ret;
	
	POP_OBJREF( fp, arg1 );

	ret = (*fn)( stp, staticself, arg1 );

	if( stp->ex || fp->ex )
		return stp->retcode;

	PUSH_INT( fp, ret );
	return 0;
}


/* static int fn(object) - (Lxxx;)I */
int wrapper_static_LI( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_static_LI fn = (fn_static_LI) gfn;
	Frame* fp = stp->fp;
	ObjRef* arg1;
	int32 ret;
	
	POP_OBJREF( fp, arg1 );

	ret = (*fn)( stp, staticself, arg1 );

	if( stp->ex || fp->ex )
		return stp->retcode;

	PUSH_INT( fp, ret );
	return 0;
}


/* static long fn(object) - (Lxxx;)J */
int wrapper_static_LJ( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_static_LJ fn = (fn_static_LJ) gfn;
	Frame* fp = stp->fp;
	ObjRef* arg1;
	u4 value[2];
	int64* ret = (int64*) value;
	
	POP_OBJREF( fp, arg1 );

	*ret = (*fn)( stp, staticself, arg1 );

	if( stp->ex || fp->ex )
		return stp->retcode;

	PUSH_LONG( fp, value );
	return 0;
}


/* static double fn(double) - (D)D */
int wrapper_static_DD( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_static_DD fn = (fn_static_DD) gfn;
	Frame* fp = stp->fp;
	u4 value[2];
	double* d = (double*) value;
	double arg1, ret;

	POP_DOUBLE( fp, value );
	arg1 = *d;

	ret = (*fn)( stp, staticself, arg1 );

	if( stp->ex || fp->ex )
		return stp->retcode;

	*d = ret;
	PUSH_DOUBLE( fp, value );
	return 0;
}


/* static object fn(object) - (Lxxx;)Lxxx; */
int wrapper_static_LL( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_static_LL fn = (fn_static_LL) gfn;
	Frame* fp = stp->fp;
	ObjRef *arg1, *ret;
	
	POP_OBJREF( fp, arg1 );

	ret = (*fn)( stp, staticself, arg1 );

	if( stp->ex || fp->ex )
		return stp->retcode;

	PUSH_OBJREF( fp, ret );
	return 0;
}


/*****************************************************************************/


typedef jint     (*fn_static_LLI)( JNIEnv* env, jclass staticself, jobject arg1, jobject arg2 );
typedef jobject  (*fn_static_LIL)( JNIEnv* env, jclass staticself, jobject arg1, jint arg2 );


/* static int fn(object,object) - (Lxxx;Lxxx;)I */
int wrapper_static_LLI( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_static_LLI fn = (fn_static_LLI) gfn;
	Frame* fp = stp->fp;
	ObjRef *arg1, *arg2;
	int32 ret;
	
	POP_OBJREF( fp, arg2 );
	POP_OBJREF( fp, arg1 );

	ret = (*fn)( stp, staticself, arg1, arg2 );

	if( stp->ex || fp->ex )
		return stp->retcode;

	PUSH_INT( fp, ret );
	return 0;
}


/* static object fn(object,int) - (Lxxx;I)Lxxx; */
int wrapper_static_LIL( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	fn_static_LIL fn = (fn_static_LIL) gfn;
	Frame* fp = stp->fp;
	ObjRef *arg1, *ret;
	int32 arg2;
	
	POP_INT( fp, arg2 );
	POP_OBJREF( fp, arg1 );

	ret = (*fn)( stp, staticself, arg1, arg2 );

	if( stp->ex || fp->ex )
		return stp->retcode;

	PUSH_OBJREF( fp, ret );
	return 0;
}


/******************************************************************************
 *
 * Special methods which can't be handled by wrapper fn's
 *
 *****************************************************************************/


/******************************************************************************
 *
 * class java.lang.Class
 *
 *****************************************************************************/

/* java.lang.Object newInstance() 
 *     throws InstantiationException, IllegalAccessException; */
int Class_newInstance( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	Frame* fp = stp->fp;
	ObjRef* self;
	ClassFile* cf;
	ObjRef* newobj;
	int retcode;

	POP_OBJREF( fp, self );
	CHECK_OBJREF( self, stp );

	cf = (ClassFile*) self->data;
	newobj = newObjectRefAndObject( cf );
	
	/* need to leave newobj on the stack AFTER <init> has been called */
	PUSH_OBJREF( fp, newobj );

	retcode = forceExecuteJavaMethod( newobj, stp, "<init>", "()V" );
	/* whereAreWe( fp ); */

	return retcode;
}


/******************************************************************************
 *
 * class java.lang.Double
 *
 *****************************************************************************/

/* static long doubleToLongBits( double d ); */
int Double_doubleToLongBits( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	Frame* fp = stp->fp;
	u4 value[2];

	POP_DOUBLE( fp, value );
	PUSH_LONG( fp, value );
	return 0;
}


/* static double longBitsToDouble( long bits ); */
int Double_longBitsToDouble( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	Frame* fp = stp->fp;
	u4 value[2];

	POP_LONG( fp, value );
	PUSH_DOUBLE( fp, value );
	return 0;
}


/******************************************************************************
 *
 * class java.lang.Float
 *
 *****************************************************************************/

/* static int floatToIntBits( float f ); */
int Float_floatToIntBits( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	Frame* fp = stp->fp;
	u4 value;

	POP_FLOAT( fp, value );
	PUSH_INT( fp, value );
	return 0;
}


/* static float intBitsToFloat( int bits ); */
int Float_intBitsToFloat( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	Frame* fp = stp->fp;
	u4 value;

	POP_INT( fp, value );
	PUSH_FLOAT( fp, value );
	return 0;
}


/******************************************************************************
 *
 * class java.io.RandomAccessFile
 *
 *****************************************************************************/

/* void open(String name, boolean writeable) throws IOException; */

int RandomAccessFile_open( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	Frame* fp = stp->fp;
	int writeable;

	POP_INT( fp, writeable );

	/* ### Cheat a little - use the existing file opening code - remember that
	 * self and the filename are handled in the called method!!! */

	if( writeable )
		return wrapper_LV( stp, staticself, (generic_fn) &Java_java_io_FileOutputStream_open );
	
	return wrapper_LV( stp, staticself, (generic_fn) &Java_java_io_FileInputStream_open );
}


/******************************************************************************
 *
 * class java.lang.Runtime
 *
 *****************************************************************************/

/* void exitInternal( int status ) */
int Runtime_exitInternal( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	wrapper_IV( stp, staticself, (generic_fn) &Java_java_lang_Runtime_exitInternal );
	return 1; /* exit */
}


/******************************************************************************
 *
 * class java.lang.System
 *
 *****************************************************************************/

/* static void arraycopy( java.lang.Object src, int src_position,
 *                        java.lang.Object dst, int dst_position,
 *                        int length );
 */
int System_arraycopy( Stack* stp, ObjRef* staticself, generic_fn gfn )
{
	Frame* fp = stp->fp;
	ObjRef *src, *dst;
	u4 src_position, dst_position, length;

	POP_INT( fp, length  );
	POP_INT( fp, dst_position );
	POP_OBJREF( fp, dst );       /* ### POP_ARRAY */
	POP_INT( fp, src_position );
	POP_OBJREF( fp, src );       /* ### POP_ARRAY */

	if( !src || !dst )
		return newCompleteException( stp, NullPointerException, "" );

	if( ((src->type == 0) && !src->cf ) || ((dst->type == 0) && !dst->cf ) )
		return newCompleteException( stp, InternalError, "Corrupt array reference" );

	Java_java_lang_System_arraycopy( stp, staticself, src, src_position, dst, dst_position, length );

	if( stp->ex || fp->ex )
		return stp->retcode;

	return 0;
}


/*****************************************************************************/


NativeMethodStr methods[] =
{
	{ "java/lang/Class",			"forName",				"(Ljava/lang/String;)Ljava/lang/Class;", &wrapper_static_LL, (generic_fn) &Java_java_lang_Class_forName },
	{ "java/lang/Class",			"getClassLoader",		"()Ljava/lang/ClassLoader;",			 &wrapper_L, (generic_fn) &Java_java_lang_Class_getClassLoader },
	{ "java/lang/Class",			"getInterfaces",		"()[Ljava/lang/Class;",					 &wrapper_L, (generic_fn) &Java_java_lang_Class_getInterfaces },
	{ "java/lang/Class",			"getName",				"()Ljava/lang/String;",					 &wrapper_L, (generic_fn) &Java_java_lang_Class_getName },
	{ "java/lang/Class",			"getPrimitiveClass",	"(Ljava/lang/String;)Ljava/lang/Class;", &wrapper_static_LL, (generic_fn) &Java_java_lang_Class_getPrimitiveClass },
	{ "java/lang/Class",			"getSuperclass",		"()Ljava/lang/Class;",					 &wrapper_L, (generic_fn) &Java_java_lang_Class_getSuperclass },
	{ "java/lang/Class",			"isArray",				"()Z",									 &wrapper_Z, (generic_fn) &Java_java_lang_Class_isArray },
	{ "java/lang/Class",			"isAssignableFrom",		"(Ljava/lang/Class;)Z",					 &wrapper_LZ, (generic_fn) &Java_java_lang_Class_isAssignableFrom },
	{ "java/lang/Class",			"newInstance",			"()Ljava/lang/Object;",					 &Class_newInstance, NULL },
	{ "java/lang/ClassLoader",		"init",					"()V",									 &wrapper_V, (generic_fn) &Java_java_lang_ClassLoader_init },
	{ "java/lang/ClassLoader",		"getSystemResourceAsStream0",	"(Ljava/lang/String;)Ljava/io/InputStream;", &wrapper_static_LL, (generic_fn) &Java_java_lang_ClassLoader_getSystemResourceAsStream0 },
	{ "java/lang/ClassLoader",		"findSystemClass0",		"(Ljava/lang/String;)Ljava/lang/Class;", &wrapper_LL, (generic_fn) &Java_java_lang_ClassLoader_findSystemClass0 },
	{ "java/lang/Double",			"doubleToLongBits",		"(D)J",									 &Double_doubleToLongBits, NULL },
	{ "java/lang/Double",			"longBitsToDouble",		"(J)D",									 &Double_longBitsToDouble, NULL },
	{ "java/io/File",				"canRead0",				"()Z",									 &wrapper_Z, (generic_fn) &Java_java_io_File_canRead0 },
	{ "java/io/File",				"delete0",				"()Z",									 &wrapper_Z, (generic_fn) &Java_java_io_File_delete0 },
	{ "java/io/File",				"exists0",				"()Z",									 &wrapper_Z, (generic_fn) &Java_java_io_File_exists0 },
	{ "java/io/File",				"isAbsolute",			"()Z",									 &wrapper_Z, (generic_fn) &Java_java_io_File_isAbsolute },
	{ "java/io/File",				"isDirectory0",			"()Z",									 &wrapper_Z, (generic_fn) &Java_java_io_File_isDirectory0 },
	{ "java/io/File",				"isFile0",				"()Z",									 &wrapper_Z, (generic_fn) &Java_java_io_File_isFile0 },
	{ "java/io/File",				"lastModified0",		"()J",									 &wrapper_J, (generic_fn) &Java_java_io_File_lastModified0 },
	{ "java/io/File",				"list0",				"()[Ljava/lang/String;",				 &wrapper_L, (generic_fn) &Java_java_io_File_list0 },
	{ "java/io/File",				"mkdir0",				"()Z",									 &wrapper_Z, (generic_fn) &Java_java_io_File_mkdir0 },
	{ "java/io/FileDescriptor",		"initSystemFD",			"(Ljava/io/FileDescriptor;I)Ljava/io/FileDescriptor;", &wrapper_static_LIL, (generic_fn) &Java_java_io_FileDescriptor_initSystemFD },
	{ "java/io/FileInputStream",	"available",			"()I",									 &wrapper_I, (generic_fn) &Java_java_io_FileInputStream_available },
	{ "java/io/FileInputStream",	"open",					"(Ljava/lang/String;)V",				 &wrapper_LV, (generic_fn) &Java_java_io_FileInputStream_open },
	{ "java/io/FileInputStream",	"read",					"()I",									 &wrapper_I, (generic_fn) &Java_java_io_FileInputStream_read },
	{ "java/io/FileInputStream",	"readBytes",			"([BII)I",								 &wrapper_LIII, (generic_fn) &Java_java_io_FileInputStream_readBytes },
	{ "java/io/FileInputStream",	"skip",					"(J)J",									 &wrapper_JJ, (generic_fn) &Java_java_io_FileInputStream_skip },
	{ "java/io/FileInputStream",	"close",				"()V",									 &wrapper_V, (generic_fn) &Java_java_io_FileInputStream_close },
	{ "java/io/FileOutputStream",	"open",					"(Ljava/lang/String;)V",				 &wrapper_LV, (generic_fn) &Java_java_io_FileOutputStream_open },
	{ "java/io/FileOutputStream",	"openAppend",			"(Ljava/lang/String;)V",				 &wrapper_LV, (generic_fn) &Java_java_io_FileOutputStream_openAppend },
	{ "java/io/FileOutputStream",	"write",				"(I)V",									 &wrapper_IV, (generic_fn) &Java_java_io_FileOutputStream_write },
	{ "java/io/FileOutputStream",	"writeBytes",			"([BII)V",								 &wrapper_LIIV, (generic_fn) &Java_java_io_FileOutputStream_writeBytes },
	{ "java/io/FileOutputStream",	"close",				"()V",									 &wrapper_V, (generic_fn) &Java_java_io_FileOutputStream_close },
	{ "java/lang/Float",			"intBitsToFloat",		"(I)F",									 &Float_intBitsToFloat, NULL },
	{ "java/lang/Float",			"floatToIntBits",		"(F)I",									 &Float_floatToIntBits, NULL },
	{ "java/lang/Math",				"floor",				"(D)D",									 &wrapper_static_DD, (generic_fn) &Java_java_lang_Math_floor },
	{ "java/lang/Object",			"clone",				"()Ljava/lang/Object;",					 &wrapper_L, (generic_fn) &Java_java_lang_Object_clone },
	{ "java/lang/Object",			"getClass",				"()Ljava/lang/Class;",					 &wrapper_L, (generic_fn) &Java_java_lang_Object_getClass },
	{ "java/lang/Object",			"hashCode",				"()I",									 &wrapper_I, (generic_fn) &Java_java_lang_Object_hashCode },
	{ "java/lang/Object",			"notifyAll",			"()V",									 &wrapper_V, (generic_fn) &Java_java_lang_Object_notifyAll },
	{ "java/lang/Object",			"wait",					"(J)V",									 &wrapper_JV, (generic_fn) &Java_java_lang_Object_wait },
	{ "java/io/ObjectInputStream",  "loadClass0",			"(Ljava/lang/Class;Ljava/lang/String;)Ljava/lang/Class;", &wrapper_LLL, (generic_fn) &Java_java_io_ObjectInputStream_loadClass0 },
	{ "java/io/ObjectOutputStream", "getRefHashCode",		"(Ljava/lang/Object;)I",				 &wrapper_static_LI, (generic_fn) &Java_java_io_ObjectOutputStream_getRefHashCode },
	{ "java/io/ObjectOutputStream", "outputClassFields",	"(Ljava/lang/Object;Ljava/lang/Class;[I)V", &wrapper_LLLV, (generic_fn) &Java_java_io_ObjectOutputStream_outputClassFields },
	{ "java/io/ObjectStreamClass", "getClassAccess",		"(Ljava/lang/Class;)I",					 &wrapper_static_LI, (generic_fn) &Java_java_io_ObjectStreamClass_getClassAccess },
	{ "java/io/ObjectStreamClass", "getFields0",			"(Ljava/lang/Class;)[Ljava/io/ObjectStreamField;", &wrapper_LL, (generic_fn) &Java_java_io_ObjectStreamClass_getFields0 },
	{ "java/io/ObjectStreamClass", "getFieldAccess",		"(Ljava/lang/Class;Ljava/lang/String;)I", &wrapper_static_LLI, (generic_fn) &Java_java_io_ObjectStreamClass_getFieldAccess },
	{ "java/io/ObjectStreamClass", "getFieldSignatures",	"(Ljava/lang/Class;)[Ljava/lang/String;", &wrapper_static_LL, (generic_fn) &Java_java_io_ObjectStreamClass_getFieldSignatures },
	{ "java/io/ObjectStreamClass", "getMethodAccess",		"(Ljava/lang/Class;Ljava/lang/String;)I", &wrapper_static_LLI, (generic_fn) &Java_java_io_ObjectStreamClass_getMethodAccess },
	{ "java/io/ObjectStreamClass", "getMethodSignatures",	"(Ljava/lang/Class;)[Ljava/lang/String;", &wrapper_static_LL, (generic_fn) &Java_java_io_ObjectStreamClass_getMethodSignatures },
	{ "java/io/ObjectStreamClass", "getSerialVersionUID",	"(Ljava/lang/Class;)J",					 &wrapper_static_LJ, (generic_fn) &Java_java_io_ObjectStreamClass_getSerialVersionUID },
	{ "java/io/ObjectStreamClass", "hasWriteObject",		"(Ljava/lang/Class;)Z",					 &wrapper_static_LZ, (generic_fn) &Java_java_io_ObjectStreamClass_hasWriteObject },
	{ "java/io/RandomAccessFile",	"open",					"(Ljava/lang/String;Z)V",				 &RandomAccessFile_open, NULL },
	{ "java/lang/Runtime",			"buildLibName",			"(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;", &wrapper_LLL, (generic_fn) &Java_java_lang_Runtime_buildLibName },
	{ "java/lang/Runtime",			"exitInternal",			"(I)V",									 &Runtime_exitInternal, NULL },
	{ "java/lang/Runtime",			"initializeLinkerInternal", "()Ljava/lang/String;",				 &wrapper_L, (generic_fn) &Java_java_lang_Runtime_initializeLinkerInternal },
	{ "java/lang/Runtime",			"loadFileInternal",		"(Ljava/lang/String;)I",				 &wrapper_LI, (generic_fn) &Java_java_lang_Runtime_loadFileInternal },
	{ "java/lang/Runtime",			"traceInstructions",	"(Z)V",									 &wrapper_ZV, (generic_fn) &Java_java_lang_Runtime_traceInstructions },
	{ "java/lang/Runtime",			"traceMethodCalls",		"(Z)V",									 &wrapper_ZV, (generic_fn) &Java_java_lang_Runtime_traceMethodCalls },
	{ "java/lang/SecurityManager",  "classLoaderDepth",		"()I",									 &wrapper_I, (generic_fn) &Java_java_lang_SecurityManager_classLoaderDepth },
	{ "java/lang/SecurityManager",  "currentClassLoader",	"()Ljava/lang/ClassLoader;",			 &wrapper_L, (generic_fn) &Java_java_lang_SecurityManager_currentClassLoader },
	{ "java/lang/SecurityManager",  "getClassContext",		"()[Ljava/lang/Class;",					 &wrapper_L, (generic_fn) &Java_java_lang_SecurityManager_getClassContext },
	{ "java/lang/System",			"arraycopy",			"(Ljava/lang/Object;ILjava/lang/Object;II)V", &System_arraycopy, NULL },
	{ "java/lang/System",			"currentTimeMillis",	"()J",									 &wrapper_static_J, (generic_fn) &Java_java_lang_System_currentTimeMillis },
	{ "java/lang/System",			"initProperties",		"(Ljava/util/Properties;)Ljava/util/Properties;", &wrapper_static_LL, (generic_fn) &Java_java_lang_System_initProperties },
	{ "java/lang/System",			"setIn0",				"(Ljava/io/InputStream;)V",				 &wrapper_static_LV, (generic_fn) &Java_java_lang_System_setIn0 },
	{ "java/lang/System",			"setOut0",				"(Ljava/io/PrintStream;)V",				 &wrapper_static_LV, (generic_fn) &Java_java_lang_System_setOut0 },
	{ "java/lang/System",			"setErr0",				"(Ljava/io/PrintStream;)V",				 &wrapper_static_LV, (generic_fn) &Java_java_lang_System_setErr0 },
	{ "java/lang/Thread",			"currentThread",		"()Ljava/lang/Thread;",					 &wrapper_static_L, (generic_fn) &Java_java_lang_Thread_currentThread },
	{ "java/lang/Thread",			"setPriority0",			"(I)V",									 &wrapper_static_IV, (generic_fn) &Java_java_lang_Thread_setPriority0 },
	{ "java/lang/Thread",			"start",				"()V",									 &wrapper_V, (generic_fn) &Java_java_lang_Thread_start },
	{ "java/lang/Throwable",		"fillInStackTrace",		"()Ljava/lang/Throwable;",				 &wrapper_L, (generic_fn) &Java_java_lang_Throwable_fillInStackTrace },
	{ "java/lang/Throwable",		"printStackTrace0",		"(Ljava/lang/Object;)V",				 &wrapper_LV, (generic_fn) &Java_java_lang_Throwable_printStackTrace0 },
	{ NULL, NULL, NULL }
};


NativeMethodStr* findNativeMethod( ClassFile *cf, u2 name_index, u2 descriptor_index )
{
	int n;
	u2 class_name = ((CONSTANT_Class_info*) cf->constant_pool[cf->this_class])->name_index;

	for( n= 0; methods[n].name && methods[n].desc
		&& !(checkSimpleName( cf, class_name, methods[n].class_name )
			&& checkSimpleName( cf, name_index, methods[n].name )
			&& checkSimpleName( cf, descriptor_index, methods[n].desc )); n++ );

	if( methods[n].name )
		return &methods[n];

	return NULL;
}
