/******************************************************************************
 *
 * Filename: java_lang_System.c
 *
 * Author:   Robert Thornburrow
 * Project:  JavaVM
 * Created:  25th March 2002
 *
 * Purpose:  Native method implementations for java.lang.System.
 *
 * (c) Copyright R.H.Thornburrow, 2002
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef WIN32
#include <sys/timeb.h>   /* for ftime() stuff */
#else
#include <sys/time.h>	 /* for gettimeofday() */
#endif

#include "../engine/config.h"
#include "../engine/disasm.h"
#include "../engine/util.h"

#include "java_lang_System.h"


/** 
 * Copies an array from the specified source array, beginning at the
 * specified position, to the specified position of the destination array.
 * A subsequence of array components are copied from the source 
 * array referenced by <code>src</code> to the destination array 
 * referenced by <code>dst</code>. The number of components copied is 
 * equal to the <code>length</code> argument. The components at 
 * positions <code>srcOffset</code> through 
 * <code>srcOffset+length-1</code> in the source array are copied into 
 * positions <code>dstOffset</code> through 
 * <code>dstOffset+length-1</code>, respectively, of the destination 
 * array. 
 * <p>
 * If the <code>src</code> and <code>dst</code> arguments refer to the 
 * same array object, then the copying is performed as if the 
 * components at positions <code>srcOffset</code> through 
 * <code>srcOffset+length-1</code> were first copied to a temporary 
 * array with <code>length</code> components and then the contents of 
 * the temporary array were copied into positions 
 * <code>dstOffset</code> through <code>dstOffset+length-1</code> of the 
 * argument array. 
 * <p>
 * If any of the following is true, an 
 * <code>ArrayStoreException</code> is thrown and the destination is 
 * not modified: 
 * <ul>
 * <li>The <code>src</code> argument refers to an object that is not an 
 *     array. 
 * <li>The <code>dst</code> argument refers to an object that is not an 
 *     array. 
 * <li>The <code>src</code> argument and <code>dst</code> argument refer to 
 *     arrays whose component types are different primitive types. 
 * <li>The <code>src</code> argument refers to an array with a primitive 
 *     component type and the <code>dst</code> argument refers to an array 
 *     with a reference component type. 
 * <li>The <code>src</code> argument refers to an array with a reference 
 *     component type and the <code>dst</code> argument refers to an array 
 *     with a primitive component type. 
 * </ul>
 * <p>
 * Otherwise, if any of the following is true, an 
 * <code>ArrayIndexOutOfBoundsException</code> is 
 * thrown and the destination is not modified: 
 * <ul>
 * <li>The <code>srcOffset</code> argument is negative. 
 * <li>The <code>dstOffset</code> argument is negative. 
 * <li>The <code>length</code> argument is negative. 
 * <li><code>srcOffset+length</code> is greater than 
 *     <code>src.length</code>, the length of the source array. 
 * <li><code>dstOffset+length</code> is greater than 
 *     <code>dst.length</code>, the length of the destination array. 
 * </ul>
 * <p>
 * Otherwise, if any actual component of the source array from 
 * position <code>srcOffset</code> through 
 * <code>srcOffset+length-1</code> cannot be converted to the component 
 * type of the destination array by assignment conversion, an 
 * <code>ArrayStoreException</code> is thrown. In this case, let 
 * <b><i>k</i></b> be the smallest nonnegative integer less than 
 * length such that <code>src[srcOffset+</code><i>k</i><code>]</code> 
 * cannot be converted to the component type of the destination 
 * array; when the exception is thrown, source array components from 
 * positions <code>srcOffset</code> through
 * <code>srcOffset+</code><i>k</i><code>-1</code> 
 * will already have been copied to destination array positions 
 * <code>dstOffset</code> through
 * <code>dstOffset+</code><i>k</I><code>-1</code> and no other 
 * positions of the destination array will have been modified. 
 *
 * @param      src:      the source array.
 * @param      srcpos    start position in the source array.
 * @param      dest      the destination array.
 * @param      destpos   start position in the destination data.
 * @param      length    the number of array elements to be copied.
 * @exception  ArrayIndexOutOfBoundsException  if copying would cause
 *               access of data outside array bounds.
 * @exception  ArrayStoreException  if an element in the <code>src</code>
 *               array could not be stored into the <code>dest</code> array
 *               because of a type mismatch.
 * @since      JDK1.0
 *
 * public static native void arraycopy(Object src, int src_position,
 *                                     Object dst, int dst_position,
 *                                     int length);
 */
void Java_java_lang_System_arraycopy( JNIEnv* env, jclass staticself,
									  jobject src, jint src_position,
									  jobject dst, jint dst_position, jint length)
{
	u4 size = 0;

	if( !((src->type == dst->type)
		|| (((src->type == T_CHAR) || (src->type == T_BYTE))
			&& ((dst->type == T_CHAR) || (dst->type == T_BYTE)))) )
	{
#ifdef _DEBUG
		char stbuf[32], dtbuf[32];
		sprintTypeName( stbuf, src->type );
		sprintTypeName( dtbuf, dst->type );
		printf( "WARNING: attempting to copy array of %s to array of %s\n", stbuf, dtbuf );
#endif
		env->retcode = newCompleteException( env, InternalError, "Not yet implemented: arraycopy of different array types" );
		return;
	}

	switch( src->type )
	{
	case T_OBJREF:  size = sizeof( ObjRef* ); break;
	case T_BOOLEAN: size = 1; break;
	case T_CHAR:	size = 1; break; /* will eventually be 2 */
	case T_FLOAT:	size = 4; break;
	case T_DOUBLE:	size = 8; break;
	case T_BYTE:	size = 1; break;
	case T_SHORT:	size = 2; break;
	case T_INT:		size = 4; break;
	case T_LONG:	size = 8; break;
	default:
#ifdef _DEBUG
		{
			char buf[32];
			sprintTypeName( buf, src->type );
			printf( "WARNING: attempting to copy array of %s\n", buf );
		}
#endif
		env->retcode = newCompleteException( env, InternalError, "Bad type in arraycopy" );
		return;
	}

	if( length == 0 )
		return;

	if( (src_position >= 0) && (dst_position >= 0) && (length > 0)
		&& (src_position + length <= (jint) src->length)
		&& (dst_position + length <= (jint) dst->length) )
	{
		u1* srcpos = (u1*) src->data;
		u1* dstpos = (u1*) dst->data;
		int32 m;
		/*char sbuf[STRING_MAX], dbuf[STRING_MAX], rbuf[STRING_MAX];
		u4 n;
		
		if( verbose && ((src->type == T_CHAR) || (src->type == T_BYTE)) ) {
			u4 srclen = ( src->length < STRING_MAX ) ? src->length : (STRING_MAX - 1);
			u4 dstlen = ( dst->length < STRING_MAX ) ? dst->length : (STRING_MAX - 1);
			sbuf[srclen]='\0'; for( n = 0; n < srclen; n++ )
			{ if( srcpos[n] < 32 ) sbuf[n] = '.'; else sbuf[n] = srcpos[n]; }
			dbuf[dstlen]='\0'; for( n = 0; n < dstlen; n++ )
			{ if( dstpos[n] < 32 ) dbuf[n] = '.'; else dbuf[n] = dstpos[n]; }
		}*/

		/*if( (&srcpos[(src_position+length)*size] > &dstpos[dst_position*size])
			&& (&srcpos[(src_position+length)*size] < &dstpos[(dst_position+length)*size]) )
			printf( "WARNING: Buffers overlap!!!\n" );*/

		/* This does a backwards memcpy to  cope with overlapping buffers */
		for( m = length*size - 1; m >= 0; m-- )
			dstpos[dst_position*size + m] = srcpos[src_position*size + m];

		/*
		if( verbose && ((src->type == T_CHAR) || (src->type == T_BYTE)) ) {
			u4 dstlen = ( dst->length < STRING_MAX ) ? dst->length : (STRING_MAX - 1);
			rbuf[dstlen]='\0'; for( n = 0; n < dstlen; n++ )
			{ if( dstpos[n] < 32 ) rbuf[n] = '.'; else rbuf[n] = dstpos[n]; }
			printf( "Copied %d bytes from \"%s\" (pos %d, len %d), into \"%s\" (pos %d, len %d) giving \"%s\"\n", length, sbuf, src_position, src->length, dbuf, dst_position, dst->length, rbuf );
		}*/

		/*
		if( src->type == T_OBJREF )
		{
			for( n = 0; n < length; n++ )
				printf( "Copied src[%d] (0x%08x) as dst[%d] (0x%08x)\n", src_position+n, src->data[src_position+n], dst_position+n, dst->data[dst_position+n] );
		}*/

		return;
	}

	env->retcode = newCompleteException( env, ArrayIndexOutOfBoundsException, "" );
}


/**
 * Returns the current time in milliseconds.
 * <p>
 * See the description of the class <code>Date</code> for a discussion 
 * of slight discrepancies that may arise between "computer 
 * time" and coordinated universal time (UTC). 
 *
 * @return  the difference, measured in milliseconds, between the current
 *          time and midnight, January 1, 1970 UTC.
 * @see     java.util.Date
 * @since   JDK1.0
 *
 * public static native long currentTimeMillis();
 */
jlong Java_java_lang_System_currentTimeMillis( JNIEnv* env, jclass staticself )
{
	jlong secs, millis;

#ifdef WIN32
	struct _timeb tbuf;

	_ftime( &tbuf );
	secs = tbuf.time;
	millis = tbuf.millitm;
#else
	struct timeval tbuf;

	gettimeofday( &tbuf, NULL );
	secs = tbuf.tv_sec;
	millis = tbuf.tv_usec;
#endif

	return (secs * 1000) + millis;
}


/* Internal method */
int internal_System_addProperty( Frame* fp, ObjRef* props, const char* property, const char* value )
{
	Stack stack;
	int retcode;
	ObjRef* dummy;

	stack.ex = NULL;
	stack.fp = fp;
	stack.tail = NULL;

	PUSH_OBJREF( fp, newJavaLangString( (u1*) property, strlen( property ) ) );
	PUSH_OBJREF( fp, newJavaLangString( (u1*) value, strlen( value ) ) );

	retcode = forceExecuteVirtualJavaMethod( props, &stack, "put", "(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;", 1 );

	while( !retcode )
		retcode = executeInstruction( &stack );

	if( retcode == 2 )
	{
		fp->ex = stack.ex ? stack.ex : (stack.fp ? stack.fp->ex : NULL );
		return retcode;
	}

	POP_OBJREF( fp, dummy );
	return 0;
}


/*
 * java.version					 = Java version number
 * java.vendor					 = Java vendor-specific string
 * java.vendor.url				 = Java vendor URL
 * java.specification.name		 = Java Platform API Specification name
 * java.specification.vendor	 = Java Platform API Specification vendor-specific string
 * java.specification.version	 = Java Platform API Specification version number
 * java.vm.specification.name	 = Java Virtual Machine Specification name
 * java.vm.specification.vendor  = Java Virtual Machine Specification vendor-specific string
 * java.vm.specification.version = Java Virtual Machine Specification version number
 * java.vm.name					 = Java Virtual Machine name
 * java.vm.vendor				 = Java Virtual Machine vendor-specific string
 * java.vm.version				 = Java Virtual Machine version number
 * java.home					 = Java installation directory
 * java.class.version			 = Java class format version number
 * java.class.path				 = Java class path
 * os.name						 = Operating system name
 * os.arch						 = Operating system architecture
 * os.version					 = Operating system version</td></tr>
 * file.separator				 = File separator ("/" on UNIX)
 * path.separator				 = Path separator (":" on UNIX)
 * line.separator				 = Line separator ("\n" on UNIX)
 * user.name					 = User's account name
 * user.home					 = User's home directory
 * user.dir						 = User's current working directory
 */

typedef enum
{
        pk_text, pk_envvar
} PropKind;

typedef struct _SystemProperty
{
	const char* name;
    PropKind kind;
	const char* value;
	const char* defval;
} SystemProperty;

SystemProperty systemProperties[] =
{
	{ "java.version",                    pk_text, "1.1.1", NULL },
	{ "java.vendor",                     pk_text, "Taurus Software", NULL },
	{ "java.vendor.url",                 pk_text, "http://www.tsac.co.uk", NULL },
	{ "java.specification.name",		 pk_text, "Java Platform API Specification", NULL },
	{ "java.specification.vendor",	     pk_text, "Sun Microsystems Inc.", NULL },
	{ "java.specification.version",	     pk_text, "1.1", NULL },
	{ "java.vm.name",					 pk_text, "Taurus JavaVM", NULL },
	{ "java.vm.specification.name",	     pk_text, "Java Virtual Machine Specification", NULL },
	{ "java.vm.specification.vendor",    pk_text, "Sun Microsystems Inc.", NULL },
	{ "java.vm.specification.version",   pk_text, "1.0", NULL },
	{ "java.vm.vendor",				     pk_text, "Taurus Software", NULL },
	{ "java.vm.version",				 pk_text, JAVAVM_VERSION, NULL },
#ifdef WIN32
	{ "java.home",           pk_envvar, "JAVA_HOME", "D:\\Projects\\javavm" },
#else
#ifdef USES_ABSOLUTE_FILENAMES
	{ "java.home",           pk_envvar, "JAVA_HOME", ABSOLUTE_FILENAME_ROOT },
#else
	{ "java.home",           pk_envvar, "JAVA_HOME", "/export/Projects/javavm" },
#endif
#endif /* WIN32 */
	{ "java.class.version",  pk_text,   "45.3", NULL },
#ifdef USES_ABSOLUTE_FILENAMES
	{ "java.class.path",     pk_text,   ABSOLUTE_FILENAME_ROOT, NULL },
#else
	{ "java.class.path",     pk_text,   ".", NULL },
#endif
	{ "awt.toolkit",         pk_text,   "sun.awt.windows.WToolkit", NULL },
	{ "file.encoding.pkg",   pk_text,   "sun.io", NULL },
	/*{ "file.encoding",	   pk_text,	  "Cp1252", NULL },*/
	{ "line.separator",      pk_text,   LINE_SEPARATOR, NULL },
	{ "file.separator",      pk_text,   FILE_SEPARATOR, NULL },
	{ "os.arch",			 pk_text,	OS_ARCH, NULL },
	{ "os.name",			 pk_text,	OS_NAME, NULL },
	{ "path.separator",      pk_text,   PATH_SEPARATOR, NULL },
#ifdef WIN32
	{ "user.home",           pk_envvar, "HOME", "D:\\Users\\Robert" },
	{ "user.dir",            pk_envvar, "PWD", "D:\\Projects\\javavm\\lib" },
#else
#ifdef USES_ABSOLUTE_FILENAMES
	{ "user.home",           pk_envvar, "HOME", ABSOLUTE_FILENAME_ROOT },
	{ "user.dir",            pk_envvar, "PWD", "." },
#else
	{ "user.home",           pk_envvar, "HOME", "/home/robert" },
	{ "user.dir",            pk_envvar, "PWD", "/export/Projects/javavm" },
#endif
#endif /* WIN32 */
	{ "user.name",           pk_envvar, "USER", "robert" },
	{ NULL, pk_text, NULL, NULL }
};


/**
 * private static native Properties initProperties(Properties props);
 */
jobject Java_java_lang_System_initProperties( JNIEnv* env, jclass staticself, jobject props )
{
	SystemProperty* add;
	env->retcode = 0;

	for( add = &systemProperties[0]; !env->retcode && add->name; add++ )
	{
	    if( add->kind == pk_envvar )
	    {
	        const char *value;

	        if( (value = getenv( add->value )) )
			{
				/*printf( "ENVIRONMENT: %s = \"%s\"\n", add->name, value );*/
				env->retcode = internal_System_addProperty( env->fp, props, add->name, value );
			}
			else
				env->retcode = internal_System_addProperty( env->fp, props, add->name, add->defval );
	    }
	    else
	    {
			env->retcode = internal_System_addProperty( env->fp, props, add->name, add->value );
	    }
	}

	return props;
}


/**
 * private static native void setIn0(InputStream in);
 */
void Java_java_lang_System_setIn0( JNIEnv* env, jclass staticself, jobject in )
{
	((ClassFile*) staticself->data)->static_data[0] = (u4) in;
}


/**
 * private static native void setOut0(PrintStream out);
 */
void Java_java_lang_System_setOut0( JNIEnv* env, jclass staticself, jobject out )
{
	((ClassFile*) staticself->data)->static_data[1] = (u4) out;
}

/**
 * private static native void setErr0(PrintStream err);
 */
void Java_java_lang_System_setErr0( JNIEnv* env, jclass staticself, jobject err )
{
	((ClassFile*) staticself->data)->static_data[2] = (u4) err;
}
